;nyquist plug-in
;version 3
;type generate
;name "Append Import..."
;action "Importing, please wait..."
;info "by Steve Daulton (www.easyspacepro.com). GPL v2.\n* A stereo track must be selected for stereo output.\n* Supported file types: .WAV .AIFF .AU.\n"

;; AppendImport.ny by Steve Daulton. Aug 2011.
;; Released under terms of the GNU General Public License version 2:
;; http://www.gnu.org/licenses/old-licenses/gpl-2.0.html

;control viewhelp "View Help" choice "No,Usage,Label Help,Limitations" 0
;control ImportMode "Import Mode" choice "Sequential,Alternate Channels" 0  
;control output "Output" choice "Mono,Stereo,Labels" 0
;control ImportFolder "Import Folder" string   

;; Globals
(setq total-dur 0)
(setq labels "")
(setq id "Append_Import_by_Steve_Daulton_12-09-2012")

(defun help ()
 (setq helpfile (format nil
"Append Import will look in the 'Import Folder' for
'filelist.txt'. If found, it will try to read each
audio file listed in 'filelist.txt' and ignores any
audio files in the Import Folder. If not found, all
supported audio files in the Import Folder will be
loaded in alphanumeric order.\n
The default Import Folder is:
\"~a\".\n
Supported file types .WAV .AIFF and .AU must have
valid file extensions. If used, 'filelist.txt' must
contain only a list of uncompressed audio files. 
Each file must be on a new line and must be either:
1) A file in the 'Import Folder'
2) A fully qualified file name.\n
Import Mode:
'Sequential' will append the files end-to-end.
'Alternate Channels' will assume that the first file
is intended for the left channel, the next file for
the right, and so on.\n
To see a list of imported file names, click the
Debug button." *default-sf-dir*))
	(format T "~a" helpfile)
	(format nil "~a" helpfile))

(defun label-help ()
 (setq helpfile (format nil
"When importing files sequentially, the files are
'merged' into one track with no split marks. Sometimes
it is useful to be able to see where each track starts
and ends, so the plug-in provides the ability to create
labels for each imported file.~%
To create labels for each file AFTER successfully
running Sequential Import, run the plug-in again with
'Output' set to 'Labels'.~%
The labels will be retained by the plug-in for use at
any time after running Append Import in 'Sequential'
mode for the duration of the Audacity session. Running
Append Import again in Sequential mode will replace
the labels stored by the plug-in with labels for the
most recent import.~%~%
Tip: To view a help screen in a form that can be easily
copied, use the Debug button with 'View Help' enabled.~%
Remember to deselect Help before running the plug-in."))
	(format T "~a" helpfile)
	(format nil "~a" helpfile))

(defun limitations ()
 (setq helpfile (format nil
"This plug-in is not tolerant of user error. Invalid
input may cause the import to fail without warning.\n
To import files into a stereo track, a stereo track
must be selected before running Append Import.\n
Non-ascii characters are not supported for file or
folder names on Windows.\n
The computer must have enough RAM for the entire
import. Attempting to import more audio than can fit
in the available memory will cause Audacity to freeze
and/or crash.\n
Progress bar and Remaining Time are non-functional.\n
Errors in the 'filelist.txt' file (if used) may
return a 'Nyquist did not return audio' error.
Using the Debug button to run the plug-in may give
some clue of what went wrong.\n
Imported files will be resampled to the current
track sample rate."))
	(format T "~a" helpfile)
	(format nil "~a" helpfile))


(defun convert (s-in)
	;get sample rate of s-in
	(setq s-in-srate 
		(if (arrayp s-in)
        (snd-srate (aref s-in 0))
        (snd-srate s-in)))
	(if (= output 0)
      ;mono out
      (when (arrayp s-in)
        (setq s-in (mult 0.5 (sim (aref s-in 0)(aref s-in 1)))))
      ;stereo out
      (when (not (arrayp s-in))
        (setq s-in (vector s-in s-in))))
	;convert sample rate if necessary
	(if (/= s-in-srate *sound-srate*)
      (resample s-in *sound-srate*) s-in))

(defun test-ext (fnm); returns T for valid files
  (cond
    ((and (>= (length fnm) 5)
          (string-equal (subseq fnm (- (length fnm) 5)) ".aiff"))
      T)
    ((and (>= (length fnm) 4)
          (or (string-equal (subseq fnm (- (length fnm) 4)) ".wav")
              (string-equal (subseq fnm (- (length fnm) 4)) ".aif")))
      T)
    ((and (>= (length fnm) 3)	
          (string-equal (subseq fnm (- (length fnm) 3)) ".au"))
      T)))

;sort list
(defun sort (testl &aux newlist)
	(dotimes (j (length testl) newlist)	
		(let ((out	""))
			(dolist (i testl newlist)
				(setq out (if (string-greaterp out i) out i)))
			(setq newlist (cons out newlist))
			(setq testl (remove out testl)))))

;get all files in path in alphanumeric order
(defun get-all ()
	(let (outlist (flst  (listdir *default-sf-dir*)))
		(dotimes (i (length flst))
			(if (test-ext (nth i flst))
          (setq outlist (cons (nth i flst) outlist))))
	 	(sort outlist)))

;print list of files
(defun pflist (fnamelist)
  (let ((result ""))
    (dolist (fname fnamelist)
      (setf result (format nil "~a~a~%" result fname)))
    result))

;split list alternate left/right
(defun split-alt (full-list)
	(let ((list-len (* 2 (/ (length full-list) 2)))
        Llist Rlist)
    (setf result
      (do ((i 0 (setq i (+ 2 i))))
          ; keep left list in reverse so that we can add a final value if required.
          ((= i list-len)(vector Llist (reverse Rlist)))
        (setf Llist (cons (nth i full-list) Llist))
        (setf Rlist (cons (nth (1+ i) full-list) Rlist))))
    (if (> (length full-list) list-len)   ; odd number of files
        (progn
          (setf (aref result 0)(cons (car (last full-list))(aref result 0)))
          (setf (aref result 0)(reverse (aref result 0))))
        (setf (aref result 0)(reverse (aref result 0))))
    result))

;;; make property symbol 'unique'.
(defun property (p-name)
  (read (make-string-input-stream
    (format nil "~a~a" id p-name))))

;;; evaluate string as an expression
(defun string-eval (str)
  (eval (read (make-string-input-stream str))))

;;; Length of mono/stereo sound in seconds
;;; Asumes both channels the same length
(defun ms-length (sig)
  (if (arrayp sig)
      (let ((sr (snd-srate (aref sig 0))))
        (/ (snd-length (aref sig 0) ny:all) sr))
      (let ((sr (snd-srate sig)))
        (/ (snd-length sig ny:all) sr))))

;;; Read files and note duration
(defun s-read-add (fname)
  (let* ((new-s (s-read fname))
         (new-len (ms-length new-s)))
    (setq labels (format nil "~a(list ~a ~a \"~a\")"
      labels 
      total-dur
      (+ total-dur new-len)
      fname))
    (setq total-dur (+ total-dur new-len))
    new-s))

;;; Import audio files
(defun import ()
	;set the import folder
	(setq *default-sf-dir* 
    (if (> (length (string-trim " " ImportFolder)) 0) 
        ImportFolder *default-sf-dir*))
	;ensure that path ends with separator
	(setq *default-sf-dir* 
		(strcat
			(string-right-trim " \\/" *default-sf-dir*)
				(string *file-separator*)))
	(setq filelist (strcat *default-sf-dir* "filelist.txt"))
  ;check that input folder exists
	(if (listdir *default-sf-dir*)
      (progn
        ;get file list
        (setq flist
          ;if filelist.txt is found
          (if (setq fp (open filelist :direction :input))
              (do ((line (read-line fp)(setq line(read-line fp))) flist)
                  ((not line) (reverse flist))
                (when (test-ext line)
                  (setq flist (cons line flist))))
              ;ELSE get all files in path
              (get-all)))
        ;print file names to debug window
        (if (> (length flist) 0)
            (format t "Import folder:~%~a~%~%Importing files:~%~a" 
              *default-sf-dir* 
              (pflist flist))
            (format t "No files found in:~%~a" 
              *default-sf-dir*))
        ;get sounds from file list
        (setf slist (mapcar #'s-read-add flist))
        (if (= ImportMode 1); Alternate Channels
            (progn 
              ;make imports to mono
              (setq output 0)
              (setf slist (mapcar #'convert slist))
              ;put left / right lists into an array
              (setq slist (split-alt slist))
              ;; create stereo sound
              ;; Audacity 2.0.2 truncates to the shortest channel
              (setf output (vector
                (seqrep (i (length (aref slist 0)))(cue (nth i (aref slist 0))))
                (seqrep (i (length (aref slist 1)))(cue (nth i (aref slist 1))))))
              (let* ((dur0 (snd-length (aref output 0) ny:all))
                     (dur1 (snd-length (aref output 1) ny:all))
                     (dur (/ (max dur0 dur1) *sound-srate*)))
                (if (= dur 0)
                    output
                    (sum (vector (s-rest dur)(s-rest dur)) output))))





            (progn ; ELSE Sequential Append		
              ;convert sounds if necessary
              (setf slist (mapcar #'convert slist))
              ;; put label code as a property of *scratch*
              (setq labels (string-eval (format nil "(list ~a)" labels)))
              (putprop '*scratch* labels (property "labels"))
              ;output sounds from list
              (seqrep (i (length slist))(cue (nth i slist))))))      
      ;OR output error
      (format nil "Error.~%Directory \"~a\" does not exist.~%" *default-sf-dir*)))

(defun make-labels ()
  (setf labels (get '*scratch* (property "labels")))
  (if labels
      (progn
        (format t "~a labels will be created." (length labels))
        labels)
      (format nil "* NO LABELS AVAILABLE *~%~%~a" (label-help))))

;If View Help
(case viewhelp
      (1 (help))
      (2 (label-help))
      (3 (limitations))
      ;ELSE run the plug-in
      (T (if (< output 2)
             (import)
             (make-labels))))
